#!/bin/bash

# parameters
# $1 which fold reducer to use
# $2 which verifier to use
# $3 condition file 
# $4 architecture (32 or 64)
# $5 property file 
# $6 program file

if [[ $1 == "--version" ]]
then
  echo "Fold-reducer based conditional verifiers 1.0"
  exit
fi

if [[ ($# < 5) ]]
then
  echo "Error: Too few arguments"
  exit
fi

SCRIPT="$(readlink -f "$0")"
DIR_OF_SCRIPT="$(readlink -f "$(dirname "$SCRIPT")")"


COND="$(readlink -f "$3")"
PROP="$(readlink -f "$5")"
PROG="$(readlink -f "$6")"



# build residual program
# at the moment we use CPAchecker to generate the residual program
 
RESIDPROGNAME="residual_program.c"

echo "Generate residual program"

cd "$DIR_OF_SCRIPT/CPAchecker/"

if [[ -e "output/$RESIDPROGNAME" ]]
then
  rm output/$RESIDPROGNAME
fi

    # generate residual program
    cmd="scripts/cpa.sh -residualProgramGenerator -$4 -spec $PROP $PROG -setprop residualprogram.file=$RESIDPROGNAME -setprop AssumptionAutomaton.cpa.automaton.inputFile=$COND -heap 10000M -stack 800M -benchmark -stats"

case "$1" in
     "cfa") ;&
     "CFA_Reducer")
       cmd="$cmd -setprop residualprogram.strategy=CONDITION_PLUS_FOLD -setprop residualprogram.folderType=CFA"
       ;;
       
     "lh") ;&
     "LH_Reducer")
       cmd="$cmd -setprop residualprogram.strategy=CONDITION_PLUS_FOLD -setprop residualprogram.folderType=LOOP_ALWAYS"
       ;;
       
     "lhb") ;&
     "LHB_Reducer")
       cmd="$cmd -setprop residualprogram.strategy=CONDITION_PLUS_FOLD -setprop residualprogram.folderType=LOOP_BOUND -setprop residualprogram.unrollBound=10"
       ;;

     "lhc") ;&
     "LHC_Reducer")
       cmd="$cmd -setprop residualprogram.strategy=CONDITION_PLUS_FOLD -setprop residualprogram.folderType=LOOP_SAME_CONTEXT"
       ;;
       
     "lhbc") ;&
     "LHBC_Reducer")
       cmd="$cmd -setprop residualprogram.strategy=CONDITION_PLUS_FOLD -setprop residualprogram.folderType=LOOP_BOUND_SAME_CONTEXT -setprop residualprogram.unrollBound=10"
       ;;
     
     "nlh") ;&
     "NLH_Reducer")
       cmd="$cmd -setprop residualprogram.strategy=CONDITION_PLUS_FOLD -setprop residualprogram.folderType=FOLD_EXCEPT_LOOPS"
       ;;
         
     "sep") ;&
     "SEP_Reducer")
       cmd="$cmd -setprop residualprogram.strategy=CONDITION"	
       ;;

     *)
       echo "Error: Unknown fold-reducer $1"
       exit 
  esac
  
    # echo "$cmd"
    genResult=`$cmd`
    echo "$genResult"

    if [[ $genResult != *"Finished."* ]]
    then
       echo "ERROR: Residual program generation did not finish normally"
       exit
    fi

  # check that everything went well before we continue, i.e.,
  # check that residual program exists
  if [[ -f "output/$RESIDPROGNAME" ]]
  then
    RESIDPROG="$DIR_OF_SCRIPT/CPAchecker/output/$RESIDPROGNAME" 
  else
    echo "ERROR: Cannot find generated residual program"
    exit
  fi


  # continue with verification
  echo "Verify generated residual program"
    
  case "$2" in
      
     "cpaSeq")
       cd "$DIR_OF_SCRIPT/cpa-seq"
       cmd="scripts/cpa.sh -svcomp20 -spec $PROP $RESIDPROG -heap 10000M -stats -stack 800M -benchmark -timelimit 900s -$4"
       ;;
       
     "esbmc")
       cd "$DIR_OF_SCRIPT/esbmc/"
       cmd="./esbmc-wrapper.py -p $PROP -s kinduction --arch $4 $RESIDPROG"
       ;;
     
     "symbiotic")
       cd "$DIR_OF_SCRIPT/symbiotic/"
       if [[ $4 == "32" ]]
       then
         cmd="./bin/symbiotic --prp=$PROP --witness witness.graphml --sv-comp --$4 $RESIDPROG"
       else
         cmd="./bin/symbiotic --prp=$PROP --witness witness.graphml --sv-comp $RESIDPROG"
       fi
       ;;

     "veriabs")
       cd "$DIR_OF_SCRIPT/VeriAbs/"
       cmd="./scripts/veriabs -${4} --property-file $PROP $RESIDPROG "
       ;;

     "kInduction")
       cd "$DIR_OF_SCRIPT/CPAchecker"
       cmd="scripts/cpa.sh -kInduction -spec $PROP $RESIDPROG -heap 10000M -stack 800M -benchmark -stats -$4"
       ;;

     "value")
       cd "$DIR_OF_SCRIPT/CPAchecker"
       cmd="scripts/cpa.sh -valueAnalysis -spec $PROP $RESIDPROG -heap 10000M -stack 800M -stats -benchmark -$4"
       ;;

     "predicate")
       cd "$DIR_OF_SCRIPT/CPAchecker"
       cmd="scripts/cpa.sh -predicateAnalysis -spec $PROP $RESIDPROG -heap 10000M -stack 800M -stats -benchmark -$4"
       ;;

       *)
       echo "Error: Unknown verification tool" 
       exit
  esac

       # echo "$cmd"
       verifierResult=`$cmd 2>&1`
       echo "$verifierResult"

#evaluate result
  case "$2" in
     "cpaSeq") ;&
     "kInduction") ;&
     "value") ;&
     "predicate")
       if [[ $verifierResult == *"Verification result: FALSE."* ]]
       then
         echo "CMC verification result: VIOLATION"
       elif [[ $verifierResult == *"Verification result: TRUE."* ]]
       then
         echo "CMC verification result: CORRECT"
       else
         echo "CMC verification result: UNKNOWN" 
       fi	
       ;;     
     "esbmc")
       if [[ $verifierResult == *"FALSE_REACH"* ]]
       then
         echo "CMC verification result: VIOLATION"
       elif [[ $verifierResult == *"TRUE"* ]]
       then
         echo "CMC verification result: CORRECT"
       else
         echo "CMC verification result: UNKNOWN" 
       fi
       ;;
     "symbiotic")
       if [[ $verifierResult == *"RESULT: false"* ]]
       then
         echo "CMC verification result: VIOLATION"
       elif [[ $verifierResult == *"RESULT: true"* ]]
       then
         echo "CMC verification result: CORRECT"
       else
         echo "CMC verification result: UNKNOWN" 
       fi
       ;;
     "veriabs")
       if [[ $verifierResult == *"VERIABS_VERIFICATION_FAILED"* ]]
       then
         echo "CMC verification result: VIOLATION"
       elif [[ $verifierResult == *"VERIABS_VERIFICATION_SUCCESSFUL"* ]]
       then
         echo "CMC verification result: CORRECT"
       else
         echo "CMC verification result: UNKNOWN" 
       fi
       ;;
     *)
       echo "CMC verification result: UNKNOWN"
  esac       
